/***********************************************************
Copyright 1989 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/* $XConsortium: mispansZ.c,v 1.8 94/01/07 15:11:11 takamori Exp $ */

#include "misc.h"
#include "pixmapstr.h"
#include "gcstruct.h"
#include "mispansZ.h"
#include "ddpex.h"
#include "ddpex3.h"

/*

These routines maintain lists of Spans, in order to implement the
``touch-each-pixel-once'' rules of wide lines and arcs.

Written by Joel McCormack, Summer 1989.

XORSpans effectively performs a "XOR" on two spans, returning the unique
elements of the span, and is used to implement the FillAreaSets primitive.
XORSpans procedure written by Daiji Takamori, Fall 1993.

*/
#define _DEBUG_OFF_
#define _DONT_IGNORE_PROC_
#define _BOUNDS_CHECKING_ON_
#define _RANGE_CHECKING_ON_

Bool miInsertSpansZ(); 
Bool miDeleteSpanZ ();

void miXORSpansZ(main, addon)
    ZSpans     *main, addon;
{
    int        mainCount, addonCount;
    ddCoord3DS *mainPt, *addonPt;
    int        *mainWid, *addonWid;
    int        *mainDep, *addonDep;
    int        addonx2, mainx2;
    int        extra;
    ddCoord3DS  *newPt;
    int         *newWid, *newDep;
    int        index;
#ifdef _BOUNDS_CHECKING_ON_
    int        mainindex, addonindex;
#endif
#ifdef _RANGE_CHECKING_ON_
if (!main->points) index /= mainindex-mainindex;
#endif

#ifdef _IGNORE_PROC_
return;
#endif
    mainCount = main->count;
    mainPt = main->points;
    mainWid = main->widths;
    mainDep = main->depths;
    addonCount = addon.count;
    addonPt = addon.points;
    addonWid = addon.widths;
    addonDep = addon.depths;
#ifdef _BOUNDS_CHECKING_ON_
    mainindex = 0;
    addonindex = 0;
#endif
    extra = 0;
    while (addonCount && mainCount)
    {                                                   /* For each span */
        if (mainPt->y < addonPt->y)
	{                                 /* Main is on earlier scanline */
#ifdef _BOUNDS_CHECKING_ON_
mainindex++;
if (mainindex>main->count) mainindex /= addonindex-addonindex;
#endif
	    mainPt++;  mainWid++;  mainDep++;  mainCount--;
	}
	else if (addonPt->y < mainPt->y)
	{                                   /* Main is on later scanline */
	                           /* So insert addon b4 main span */
#ifdef _DEBUG_OFF_
	    if (!miInsertSpansZ(main, index = mainPt-main->points,
				addonPt, addonWid, addonDep, 1))
	      break;
#ifdef _BOUNDS_CHECKING_ON_
mainindex++;
if (mainindex>main->count) mainindex /= addonindex-addonindex;
#endif
	    mainPt = main->points + ++index; 
	    mainWid = main->widths + index;
	    mainDep = main->depths + index;
#endif
	    /* increment the addon index */
#ifdef _BOUNDS_CHECKING_ON_
addonindex++;
if (addonindex>addon.count) addonindex /= mainindex-mainindex;
#endif
	    addonPt++;  addonWid++; addonDep++;  addonCount--;
	}
        else 
	{                                               /* Same scan line */
	    mainx2 = mainPt->x + *mainWid;
	    addonx2 = addonPt->x + *addonWid;
	    if (mainx2 < addonPt->x) 
            {   /* main b4 addon; just skip to the next main */
#ifdef _BOUNDS_CHECKING_ON_
mainindex++;
if (mainindex>main->count) mainindex /= addonindex-addonindex;
#endif
		mainPt++;  mainWid++;  mainDep++;  mainCount--;
	    }
	    else if (addonx2 < mainPt->x)
	    {
#ifdef _DEBUG_OFF_
	        /* insert addon b4 main span */
	        if (!miInsertSpansZ(main, index = mainPt-main->points,
				    addonPt, addonWid, addonDep, 1))
		  break;
#ifdef _BOUNDS_CHECKING_ON_
mainindex++;
if (mainindex>main->count) mainindex /= addonindex-addonindex;
#endif
	    mainPt = main->points + ++index; 
	    mainWid = main->widths + index;
	    mainDep = main->depths + index;
#endif
#ifdef _BOUNDS_CHECKING_ON_
addonindex++;
if (addonindex>addon.count) addonindex /= mainindex-mainindex;
#endif
	        addonPt++;  addonWid++; addonDep++;  addonCount--;
	    }
	    else
	    {
	        int insert = 0;
	        ddCoord3DS tmpPt;   /* tmp is used for when a span is split */
		int tmpWid, tmpWid2, tmpDep;

		if (mainPt->x < addonPt->x)
		  {
		    /* main gets split up into main + tmp*/
		    insert = 1;
		    tmpPt.x = addonPt->x;
		    tmpPt.z = addonPt->z;
		    tmpWid = mainx2 - addonPt->x;
		    tmpDep = tmpWid / *mainWid * *mainDep;
		    tmpWid2 = *mainWid;
		    *mainWid = addonPt->x - 1 - mainPt->x;
		    *mainDep = *mainWid / (tmpWid2+1) * (*mainDep+1);
		  }
		else if (addonPt->x < mainPt->x)
		  {
		    /* addon gets split up into main + addon, tmp equals main*/
		    insert = 1;
		    tmpPt.x = mainPt->x;
		    tmpPt.z = mainPt->z;
		    tmpWid = *mainWid;
		    tmpDep = *mainDep;
		    mainPt->x = addonPt->x;
		    mainPt->z = addonPt->z;
		    *mainWid = tmpPt.x - 1 - addonPt->x;
		    *mainDep = *mainWid / *addonWid * *addonDep;
		    addonPt->x = tmpPt.x;
		    addonPt->z = tmpPt.z;
		    tmpWid2 = *addonWid;
		    *addonWid = addonx2 - tmpPt.x;
		    *addonDep = *addonWid / (tmpWid2+1) * (*addonDep+1);
		  }
		/* now addon and tmp (or main, if tmp was not defined)
		   start at the same place.  We know if tmp was defined 
		   because then insert = 1 */
		if (mainx2 > addonx2)
		  {
		    /* make span start at addonx2+1-- main is longer than
		        addon.  */
		    if (insert)
		      {		
			/* We'll need to insert tail edge of main (in tmp) */
#ifdef _DEBUG_OFF_
#ifdef _BOUNDS_CHECKING_ON_
mainindex++;
if (mainindex>main->count) mainindex /= addonindex-addonindex;
#endif
			mainPt++; mainWid++; mainDep++;
			if (!miInsertSpansZ(main, index = mainPt-main->points,
					    addonPt, addonWid, addonDep, 1))
			  break;
			mainPt = main->points + index; 
			mainWid = main->widths + index;
			mainDep = main->depths + index;
#endif			
		      }
		    /* Cut off the left side of main, keep the rest */
		    mainPt->x = addonx2 + 1;
		    mainPt->z = addonPt->z + *addonDep;
		    *mainWid = mainx2 - addonx2 - 1;
		    *mainDep = *mainWid / (*addonWid+1) * (*addonDep+1);
		    /* Increment the indices, decrement the addon count */
#ifdef _BOUNDS_CHECKING_ON_
addonindex++;
if (addonindex>addon.count) addonindex /= mainindex-mainindex;
#endif
		    addonPt++;  addonWid++;  addonDep++;  addonCount--;
		  }
		else { /* not inserting spans for sure */
		  if (addonx2 > mainx2)
		    {
		      /* make span start at mainx2+1 -- addon is longer than 
			 main */

		      /* Cut off the left side of addon, keep the rest */
		      addonPt->x = mainx2 + 1;
		      addonPt->z = mainPt->z + *mainDep;
		      *addonWid = addonx2 - mainx2 - 1;
		      *addonDep = *addonWid / (*mainWid+1) * (*mainDep+1);
		    }
		  else 
		    {  /* Overlapping on the right side */
		      /* Increment the addon indices, decrement addon count */
#ifdef _BOUNDS_CHECKING_ON_
addonindex++;
if (addonindex>addon.count) addonindex /= mainindex-mainindex;
#endif
		      addonPt++;  addonWid++;  addonDep++; addonCount--;
		    }
		  if (insert)
		    { /* Increment the indices */
#ifdef _BOUNDS_CHECKING_ON_
mainindex++;
if (mainindex>main->count) mainindex /= addonindex-addonindex;
#endif
		      mainPt++;   mainWid++;   mainDep++;
		    }
		  else
		    {	/* Overlapping on left.  Remove the span from main. */
#ifdef _DEBUG_OFF_
#ifdef _RANGE_CHECKING_ON_
if (!main->points) index /= mainindex-mainindex;
#endif
		      if (!miDeleteSpanZ(main, index = mainPt-main->points))
#ifdef _RANGE_CHECKING_ON_IGNORED_HERE_
{
if (!main->points) index /= mainindex-mainindex;
break;
}
#else
			break;
#endif
		      mainPt = main->points + index; 
		      mainWid = main->widths + index;
		      mainDep = main->depths + index;
#endif
		    }
		  mainCount--;
		}
	    }
	}
    }
    if (!main->count) return;
#ifdef _RANGE_CHECKING_ON_
if (!main->points) index /= mainindex-mainindex;
#endif
#ifdef _DEBUG_OFF_
#ifdef _RANGE_CHECKING_ON_
{
int i;
if (mainPt-main->points > main->count) i/=mainPt-mainPt;
if (!miInsertSpansZ (main, i=mainPt-main->points, 
		     addonPt, addonWid, addonDep, addonCount))
      return;
}
#else
    if (!miInsertSpansZ (main, mainPt-main->points, 
			 addonPt, addonWid, addonDep, addonCount))
      return;
#endif
#endif
} /* miDisjunctionSpansZ */

Bool miInsertSpansZ (spans,index,points,widths,depths,count)
     ZSpans *spans;
     ddCoord3DS *points;
     int *widths, *depths, count;
{
  ddCoord3DS *newPt;
  int *newWid, *newDep;
  
#ifdef _RANGE_CHECKING_ON_
if (index>spans->count) index /= count-count;
#endif
  spans->points = newPt = (ddCoord3DS*) xrealloc (spans->points, 
						  (spans->count + count) 
						  * sizeof (ddCoord3DS));
  spans->widths = newWid = (int *) xrealloc (spans->widths, 
					     (spans->count + count) 
					     * sizeof (int));
  spans->depths = newDep = (int *) xrealloc (spans->depths, 
					     (spans->count + count)
					     * sizeof (int));
  if (!newPt || !newWid || !newDep)
    return(FALSE);
#ifdef _RANGE_CHECKING_ON_
if (index>spans->count) index /= count-count;
#endif
  newPt += index; newWid += index; newDep += index;
  bcopy (newPt, newPt + count, sizeof(ddCoord3DS) * (spans->count - index));
  bcopy (newWid, newWid + count, sizeof(int) * (spans->count - index));
  bcopy (newDep, newDep + count, sizeof(int) * (spans->count - index));
  spans->count += count; 
#ifdef _RANGE_CHECKING_ON_
if (index>spans->count) index /= count-count;
#endif
  bcopy (points, newPt, sizeof(ddCoord3DS) * count);
  bcopy (widths, newWid, sizeof(int) * count);
  bcopy (depths, newDep, sizeof(int) * count);
  return(TRUE);
}

Bool miDeleteSpanZ (spans,index)
     ZSpans *spans;
     int index;
{
#ifdef _RANGE_CHECKING_ON_
if (!spans->points) index /= index-index;
#endif
  bcopy (spans->points + index + 1, spans->points + index, 
	   sizeof(ddCoord3DS) * (spans->count - index -1));
#ifdef _RANGE_CHECKING_ON_
if (!spans->points) index /= index-index;
#endif
  bcopy (spans->widths + index + 1, spans->widths + index, 
	   sizeof(int) * (spans->count - index - 1));
  bcopy (spans->depths + index + 1, spans->depths + index, 
	   sizeof(int) * (spans->count - index - 1));
  spans->count--;
  spans->points = (ddCoord3DS*) xrealloc (spans->points, 
					  spans->count * sizeof (ddCoord3DS));
  spans->widths = (int *) xrealloc (spans->widths, 
				    spans->count * sizeof (int));
  spans->depths = (int *) xrealloc (spans->depths, 
				    spans->count * sizeof (int));
#ifdef _RANGE_CHECKING_ON_IGNORED_HERE_
if (!spans->points) index /= index-index;
#endif
  if (!spans->points || !spans->widths || !spans->depths)
    return(FALSE);
  return(TRUE);
}

void miInitSpanGroupZ(spanGroup)
    ZSpanGroup *spanGroup;
{
    spanGroup->size = 0;
    spanGroup->count = 0;
    spanGroup->group = NULL;
    spanGroup->ymin = MAXSHORT;
    spanGroup->ymax = MINSHORT;
} /* InitSpanGroupZ */

#define YMIN(spans) (spans->points[0].y)
#define YMAX(spans)  (spans->points[spans->count-1].y)

void miSubtractSpansZ (spanGroup, sub)
    ZSpanGroup	*spanGroup;
    ZSpans	*sub;
{
    int		i, subCount, spansCount;
    int		ymin, ymax, xmin, xmax;
    ZSpans	*spans;
    ddCoord3DS	*subPt, *subPtNext, *spansPt, *spansPtNext;
    int		*subWid, *spansWid;
    int		*subDep, *spansDep;
    int		extra;

    ymin = YMIN(sub);
    ymax = YMAX(sub);
    spans = spanGroup->group;
    for (i = spanGroup->count; i; i--, spans++) {
	if (YMIN(spans) <= ymax && ymin <= YMAX(spans)) {
	    subCount = sub->count;
	    subPt = sub->points;
	    subWid = sub->widths;
	    subDep = sub->depths;
	    spansCount = spans->count;
	    spansPt = spans->points;
	    spansWid = spans->widths;
	    spansDep = spans->depths;
	    extra = 0;
	    for (;;)
 	    {
		while (spansCount && spansPt->y < subPt->y)
		{
		    spansPt++;  spansWid++;  spansDep++;  spansCount--;
		}
		if (!spansCount)
		    break;
		while (subCount && subPt->y < spansPt->y)
		{
		    subPt++;	subWid++;  subDep++;  subCount--;
		}
		if (!subCount)
		    break;
		if (subPt->y == spansPt->y)
		{
		    xmin = subPt->x;
		    xmax = xmin + *subWid;
		    if (xmin >= spansPt->x + *spansWid || spansPt->x >= xmax)
		    {
			;
		    }
		    else if (xmin <= spansPt->x)
		    {
			if (xmax >= spansPt->x + *spansWid) 
			{
			    bcopy (spansPt + 1, spansPt, sizeof *spansPt * (spansCount - 1));
			    bcopy (spansWid + 1, spansWid, sizeof *spansWid * (spansCount - 1));
			    spansPt--;
			    spansWid--;
			    spans->count--;
			    extra++;
			}
			else 
			{
			    *spansDep = *spansDep - (((xmax - spansPt->x) * *spansDep) / *spansWid);
			    *spansWid = *spansWid - (xmax - spansPt->x);
			    spansPt->x = xmax;
			}
		    }
		    else
		    {
			if (xmax >= spansPt->x + *spansWid)
			{
			    *spansWid = xmin - spansPt->x;
			    *spansDep = ((xmin - spansPt->x) * *spansDep) / *spansWid;
			}
			else
			{
			    if (!extra) {
				ddCoord3DS  *newPt;
				int	    *newwid, *newdep;

#define EXTRA 8
				newPt = (ddCoord3DS *) xrealloc (spans->points, (spans->count + EXTRA) * sizeof (ddCoord3DS));
				if (!newPt)
				    break;
				spansPt = newPt + (spansPt - spans->points);
				spans->points = newPt;
				newwid = (int *) xrealloc (spans->widths, (spans->count + EXTRA) * sizeof (int));
				newdep = (int *) xrealloc (spans->depths, (spans->count + EXTRA) * sizeof (int));
				if (!newwid)
				    break;
				spansWid = newwid + (spansWid - spans->widths);
				spansDep = newdep + (spansDep - spans->depths);
				spans->widths = newwid;
				spans->depths = newdep;
				extra = EXTRA;
			    }
			    bcopy (spansPt, spansPt + 1, sizeof *spansPt * (spansCount));
			    bcopy (spansWid, spansWid + 1, sizeof *spansWid * (spansCount));
			    bcopy (spansDep, spansDep + 1, sizeof *spansDep * (spansCount));
			    spans->count++;
			    extra--;
			    *spansDep = ((xmin - spansPt->x) * *spansDep) / *spansWid;
			    *spansWid = xmin - spansPt->x;
			    spansDep++;
			    spansWid++;
			    spansPt++;
			    *spansDep = *spansDep - (((xmax - spansPt->x) * *spansDep) / *spansWid);
			    *spansWid = *spansWid - (xmax - spansPt->x);
			    spansPt->x = xmax;
			}
		    }
		}
		spansPt++;  spansWid++; spansDep++; spansCount--;
	    }
	}
    }
}
    
void miAppendSpansZ(spanGroup, otherGroup, spans)
    ZSpanGroup   *spanGroup;
    ZSpanGroup	*otherGroup;
    ZSpans       *spans;
{
    register    int ymin, ymax;
    register    int spansCount;

    spansCount = spans->count;
    if (spansCount > 0) {
	if (spanGroup->size == spanGroup->count) {
	    spanGroup->size = (spanGroup->size + 8) * 2;
	    spanGroup->group = (ZSpans *)
		xrealloc(spanGroup->group, sizeof(ZSpans) * spanGroup->size);
	 }

	spanGroup->group[spanGroup->count] = *spans;
	(spanGroup->count)++;
	ymin = spans->points[0].y;
	if (ymin < spanGroup->ymin) spanGroup->ymin = ymin;
	ymax = spans->points[spansCount - 1].y;
	if (ymax > spanGroup->ymax) spanGroup->ymax = ymax;
	if (otherGroup &&
	    otherGroup->ymin < ymax &&
	    ymin < otherGroup->ymax)
	{
	    miSubtractSpansZ (otherGroup, spans);
	}
    }
    else
    {
	xfree (spans->points);
	xfree (spans->widths);
    }
} /* AppendSpansZ */

void miFreeSpanGroupZ(spanGroup)
    ZSpanGroup   *spanGroup;
{
    if (spanGroup->group != NULL) xfree(spanGroup->group);
}

static void QuickSortSpansXZ(points, widths, depths, numSpans)
    register ddCoord3DS     points[];
    register int	    widths[];
    register int	    depths[];
    register int	    numSpans;
{
    register int	    x;
    register int	    i, j, m;
    register ddCoord3DS     *r;

/* Always called with numSpans > 1 */
/* Sorts only by x, as all y should be the same */

#define ExchangeSpansZ(a, b)				    \
{							    \
    ddCoord3DS      tpt;				    \
    register int    tw, td;				    \
							    \
    tpt = points[a]; points[a] = points[b]; points[b] = tpt;    \
    tw = widths[a]; widths[a] = widths[b]; widths[b] = tw;  \
    td = depths[a]; depths[a] = depths[b]; depths[b] = td;  \
}

    do {
	if (numSpans < 9) {
	    /* Do insertion sort */
	    register int xprev;

	    xprev = points[0].x;
	    i = 1;
	    do { /* while i != numSpans */
		x = points[i].x;
		if (xprev > x) {
		    /* points[i] is out of order.  Move into proper location. */
		    ddCoord3DS tpt;
		    int	    tw, td, k;

		    for (j = 0; x >= points[j].x; j++) {}
		    tpt = points[i];
		    tw  = widths[i];
		    td  = depths[i];
		    for (k = i; k != j; k--) {
			points[k] = points[k-1];
			widths[k] = widths[k-1];
			depths[k] = depths[k-1];
		    }
		    points[j] = tpt;
		    widths[j] = tw;
		    depths[j] = td;
		    x = points[i].x;
		} /* if out of order */
		xprev = x;
		i++;
	    } while (i != numSpans);
	    return;
	}

	/* Choose partition element, stick in location 0 */
	m = numSpans / 2;
	if (points[m].x > points[0].x)		ExchangeSpansZ(m, 0);
	if (points[m].x > points[numSpans-1].x) ExchangeSpansZ(m, numSpans-1);
	if (points[m].x > points[0].x)		ExchangeSpansZ(m, 0);
	x = points[0].x;

        /* Partition array */
        i = 0;
        j = numSpans;
        do {
	    r = &(points[i]);
	    do {
		r++;
		i++;
            } while (i != numSpans && r->x < x);
	    r = &(points[j]);
	    do {
		r--;
		j--;
            } while (x < r->x);
            if (i < j) ExchangeSpansZ(i, j);
        } while (i < j);

        /* Move partition element back to middle */
        ExchangeSpansZ(0, j);

	/* Recurse */
        if (numSpans-j-1 > 1)
	    QuickSortSpansXZ(&points[j+1], &widths[j+1], &depths[j+1], numSpans-j-1);
        numSpans = j;
    } while (numSpans > 1);
} /* QuickSortSpansZ */


static int UniquifySpansXZ(spans, newPoints, newWidths, newDepths)
    ZSpans		    *spans;
    register ddCoord3DS     *newPoints;
    register int	    *newWidths;
    register int	    *newDepths;
{
    register int newx1, newx2, newz1, newz2, oldpt, oldptz, i, y;
    register ddCoord3DS     *oldPoints;
    register int	    *oldWidths;
    register int	    *oldDepths;
    int			    *startNewWidths;
    int			    *startNewDepths;

/* Always called with numSpans > 1 */
/* Uniquify the spans, and stash them into newPoints, newWidths, and newDepths.  Return the
   number of unique spans. */


    startNewWidths = newWidths;
    startNewDepths = newDepths;

    oldPoints = spans->points;
    oldWidths = spans->widths;
    oldDepths = spans->depths;

    y = oldPoints->y;
    newx1 = oldPoints->x;
    newx2 = newx1 + *oldWidths;
    newz1 = oldPoints->z;
    newz2 = newz1 + *oldDepths;

    for (i = spans->count-1; i != 0; i--) {
	oldPoints++;
	oldWidths++;
	oldDepths++;
	oldpt = oldPoints->x;
	oldpt = oldPoints->z;
	if (oldpt > newx2) {
	    /* Write current span, start a new one */
	    newPoints->x = newx1;
	    newPoints->y = y;
	    newPoints->z = newz1;
	    *newWidths = newx2 - newx1;
	    *newDepths = newz2 - newz1;
	    newPoints++;
	    newWidths++;
	    newDepths++;
	    newx1 = oldpt;
	    newx2 = oldpt + *oldWidths;
	    newz1 = oldptz;
	    newz2 = oldptz + *oldDepths;
	} else {
	    /* extend current span, if old extends beyond new */
	    oldpt = oldpt + *oldWidths;
	    oldptz = oldptz + *oldDepths;
	    if (oldpt > newx2) { newx2 = oldpt; newz2 = oldptz;}
	}
    } /* for */

    /* Write final span */
    newPoints->x = newx1;
    *newWidths = newx2 - newx1;
    newPoints->y = y;
    newPoints->z = newz1;
    *newDepths = newz2 - newz1;

    return (newWidths - startNewWidths) + 1;
} /* UniquifySpansXZ */

void
miDisposeSpanGroupZ (spanGroup)
    ZSpanGroup	*spanGroup;
{
    int	    i;
    ZSpans   *spans;

    for (i = 0; i < spanGroup->count; i++)
    {
	spans = spanGroup->group + i;
	xfree (spans->points);
	xfree (spans->widths);
	xfree (spans->depths);
    }
}

void miFillUniqueSpanGroupZ(zbs_ptr, pDraw, pGC, spanGroup)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr pDraw;
    GCPtr	pGC;
    ZSpanGroup   *spanGroup;
{
    register int    i;
    register ZSpans *spans;
    register ZSpans *yspans;
    register int    *ysizes;
    register int    ymin, ylength;

    /* Outgoing spans for one big call to FillSpans */
    register ddCoord3DS     *points;
    register int	    *widths, *depths;
    register int	    count;

    if (spanGroup->count == 0) return;

    if (spanGroup->count == 1) {
	/* Already should be sorted, unique */
	spans = spanGroup->group;
	ZBuffer_FillSpans(zbs_ptr, pDraw, pGC,
			  spans->count, spans->points, 
			  spans->widths, spans->depths);
	xfree(spans->points);
	xfree(spans->widths);
	xfree(spans->depths);
    }
    else
    {
	/* Yuck.  Gross.  Radix sort into y buckets, then sort x and uniquify */
	/* This seems to be the fastest thing to do.  I've tried sorting on
	   both x and y at the same time rather than creating into all those
	   y buckets, but it was somewhat slower. */

	ymin    = spanGroup->ymin;
	ylength = spanGroup->ymax - ymin + 1;

	/* Allocate Spans for y buckets */
	yspans = (ZSpans *) xalloc(ylength * sizeof(ZSpans));
	ysizes = (int *) xalloc(ylength * sizeof (int));

	if (!yspans || !ysizes)
	{
	    xfree (yspans);
	    xfree (ysizes);
	    miDisposeSpanGroupZ (spanGroup);
	}
	
	for (i = 0; i != ylength; i++) {
	    ysizes[i]        = 0;
	    yspans[i].count  = 0;
	    yspans[i].points = NULL;
	    yspans[i].widths = NULL;
	    yspans[i].depths = NULL;
	}

	/* Go through every single span and put it into the correct bucket */
	count = 0;
	for (i = 0, spans = spanGroup->group;
		i != spanGroup->count;
		i++, spans++) {
	    int		index;
	    int		j;

	    for (j = 0, points = spans->points, widths = spans->widths, depths = spans->depths;
		    j != spans->count;
		    j++, points++, widths++, depths++) {
		index = points->y - ymin;
		if (index >= 0 && index < ylength) {
		    ZSpans *newspans = &(yspans[index]);
		    if (newspans->count == ysizes[index]) {
			ddCoord3DS  *newpoints;
			int	    *newwidths, *newdepths;
			ysizes[index] = (ysizes[index] + 8) * 2;
			newpoints = (ddCoord3DS *) xrealloc(
			    newspans->points,
			    ysizes[index] * sizeof(ddCoord3DS));
			newwidths = (int *) xrealloc(
			    newspans->widths,
			    ysizes[index] * sizeof(int));
			newdepths = (int *) xrealloc(
			    newspans->depths,
			    ysizes[index] * sizeof(int));
			if (!newpoints || !newwidths || !newdepths)
			{
			    int	i;

			    for (i = 0; i < ylength; i++)
			    {
				xfree (yspans[i].points);
				xfree (yspans[i].widths);
				xfree (yspans[i].depths);
			    }
			    xfree (yspans);
			    xfree (ysizes);
			    miDisposeSpanGroupZ (spanGroup);
			    return;
			}
			newspans->points = newpoints;
			newspans->widths = newwidths;
			newspans->depths = newdepths;
		    }
		    newspans->points[newspans->count] = *points;
		    newspans->widths[newspans->count] = *widths;
		    newspans->depths[newspans->count] = *depths;
		    (newspans->count)++;
		} /* if y value of span in range */
	    } /* for j through spans */
	    count += spans->count;
	    xfree(spans->points);
	    spans->points = NULL;
	    xfree(spans->widths);
	    spans->widths = NULL;
	    xfree(spans->depths);
	    spans->depths = NULL;
	} /* for i thorough Spans */

	/* Now sort by x and uniquify each bucket into the final array */
	points = (ddCoord3DS *) xalloc(count * sizeof(ddCoord3DS));
	widths = (int *)       xalloc(count * sizeof(int));
	depths = (int *)       xalloc(count * sizeof(int));
	if (!points || !widths || !depths)
	{
	    int	i;

	    for (i = 0; i < ylength; i++)
	    {
		xfree (yspans[i].points);
		xfree (yspans[i].widths);
		xfree (yspans[i].depths);
	    }
	    xfree (points);
	    xfree (widths);
	    xfree (depths);
	    xfree (yspans);
	    xfree (ysizes);
	    return;
	}
	count = 0;
	for (i = 0; i != ylength; i++) {
	    int ycount = yspans[i].count;
	    if (ycount > 0) {
		if (ycount > 1) {
		    QuickSortSpansXZ(yspans[i].points, yspans[i].widths, yspans[i].depths, 
				     ycount);
		    count += UniquifySpansXZ
			(&(yspans[i]), &(points[count]), &(widths[count]), &(depths[count]));
		} else {
		    points[count] = yspans[i].points[0];
		    widths[count] = yspans[i].widths[0];
		    depths[count] = yspans[i].depths[0];
		    count++;
		}
		xfree(yspans[i].points);
		xfree(yspans[i].widths);
		xfree(yspans[i].depths);
	    }
	}

	ZBuffer_FillSpans(zbs_ptr, pDraw, pGC, count, points, widths, depths);
	xfree(points);
	xfree(widths);
	xfree(depths);
	xfree(yspans);
	xfree(ysizes);
    }

    spanGroup->count = 0;
    spanGroup->ymin = MAXSHORT;
    spanGroup->ymax = MINSHORT;
}


void miFillSpanGroupZ(zbs_ptr, pDraw, pGC, spanGroup)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr pDraw;
    GCPtr	pGC;
    ZSpanGroup   *spanGroup;
{
    register int    i;
    register ZSpans  *spans;

    for (i = 0, spans = spanGroup->group; i != spanGroup->count; i++, spans++) {
	ZBuffer_FillSpans(zbs_ptr, pDraw, pGC, spans->count, spans->points, spans->widths, spans->depths);
	xfree(spans->points);
	xfree(spans->widths);
	xfree(spans->depths);
    }

    spanGroup->count = 0;
    spanGroup->ymin = MAXSHORT;
    spanGroup->ymax = MINSHORT;
} /* FillSpanGroupZ */
